/** Abstract base class for Nodes and Leafs */
abstract class Tree[T] { 
 
    val content : T

    /** A higher-order generalization for tree operations.
      *                   
      * This method provides depth first tree traversal as an 
      * abstraction over different tree operations.
      * Tree operations can be implemented by creating first-class functions 
      * proc_node and proc_leaf and passing them to traverse.
      * The abstract method is implemented in Node and Leaf.                    
      */                     
    def traverse[A](proc_node: (A,A,T) => A, proc_leaf: T=>A) : A 

        override def toString = {
            val string_proc_leaf: T=>String = (x:T) => x.toString

            val string_proc_node: (String, String, T)=>String  =
                (lchild: String, rchild: String,content: T) => 
                    "("+content.toString +" "+lchild+" "+rchild+")"

            traverse(string_proc_node, string_proc_leaf)
        }
   
        def leafs : List[T] = { 

            val leaf_proc_leaf = (content: T) => List(content)
            val leaf_proc_node = 
                (lchild: List[T], rchild: List[T], content: T) =>
                    lchild ::: rchild

            traverse(leaf_proc_node, leaf_proc_leaf)
        }

      def evaluate : Int = {
            
            val eval_proc_node = (lchild: Int, rchild: Int, content: T) =>
            
                if (content=="+")
                    lchild + rchild
                else if (content=="-")
                    lchild - rchild
                else if (content=="*")
                    lchild * rchild
                else 0
                
            val eval_proc_leaf = (content: T) => content.toString.toInt

            traverse(eval_proc_node, eval_proc_leaf)
      }
}

/** A Tree with exactly two subtrees. */
class Node[T](val content: T, val left: Tree[T], val right: Tree[T]) extends Tree[T]{

    /** The traverse implementation for Node calls proc_node on the 
      * results returned by calling traverse recusively on each 
      * subtree and the content of this node. 
      */
    def traverse[A](proc_node : (A,A,T) =>A, proc_leaf: T=>A) = 
            proc_node(left.traverse(proc_node, proc_leaf), 
                     right.traverse(proc_node, proc_leaf),
                     content)
}
/** Companion object for the Node -- only used to define an apply method */
object Node {
    def apply[T](content: T, left : Tree[T], right: Tree[T]) = 
        new Node(content, left, right)
}

/** A Tree that does not have any further subtrees
  * (i.e. a single leaf node itself).
  */
class Leaf[T](val content: T) extends Tree[T] {

    /** The traverse implementation for Leaf calls proc_laf on the content of
      * the node. proc_leaf usually just converts the content into the correct
      * result type.
      */
    def traverse[A](proc_node : (A,A,T)=>A, proc_leaf: T=>A) =
        proc_leaf(content) 
}
/**  Companion object for the Leaf, only used to define an apply method */
object Leaf {
    def apply[T](content: T) = new Leaf(content) 
}    


/**
 * Main object to test tree operations. 
 */
object Part2{
  
    def main(args : Array[String]) {
        val tree : Tree[String] = 
            Node("-",
                Node("+",
                    Leaf("3"),
                    Node("*",
                        Leaf("5"),
                        Leaf("6"))),
                Leaf("7"))

        println(tree)
        // Should print (- (+ 3 (* 5 6)) 7)
        
        println(tree.leafs)  
        // Should print List(3, 5, 6, 7)

        println(tree.evaluate)
        // Should print 26       
    }
}
